# Getting Started
## Installation
### Installing python
This is platform dependent, so you need to figure out how to install python
by yourself, if it is not already installed on your system.
This will get you started:

- [https://wiki.python.org/moin/BeginnersGuide](https://wiki.python.org/moin/BeginnersGuide)  
- [https://www.python.org/downloads/](https://www.python.org/downloads/)  

In the following snippets, depending on your installation
and platform, you may have to substitute `python` with
`py` or `python3`.

### Installing the Coscine Python SDK package
#### Using pip
The Coscine Python SDK package is hosted on the
Python Package Index (PyPi). You can install and update it
and all of its dependencies via the Python Package
Manager (pip):
```sh
python -m pip install --upgrade coscine
```
The `--upgrade` option updates the package if it is already
installed on your system.

#### Using conda
This module's pypi version is mirrored in conda forge. You can
install it and all of its dependencies via Conda:
```sh
conda install -c conda-forge coscine
```
and likewise for updating it:
```sh
conda update -c conda-forge coscine
```

#### Using git
You can install the python module with a local copy of the 
source code, which you can grab directly via git:
```sh
git clone https://git.rwth-aachen.de/coscine/community-features/coscine-python-sdk.git
cd ./coscine-python-sdk
py setup.py
```
Please note that the git version may contain features not
yet released on any of the other hosting platforms. There
features might not be stable yet or provisional. In general
the git repository contains the most up to date version but
might be a little rough around the edges in certain areas. 
If you would like a stable release, refer to one of the 
git tags.

## Creating an API token
You need an API token to use the Coscine API. If you have not already,
[create a new API token](https://docs.coscine.de/en/user/token/).
Once you have an API token you are ready to use the API.

### A word of advice
The token represents sensible data and grants anyone in
possesion of it full access to your data in coscine.
Do not leak it publicly on online platforms such as github! Do
not include it within your sourcecode if you intend on
uploading that to the internet or sharing it among peers.
Take precautions and follow best practices to avoid 
corruption, theft or loss of data!

### Using the API token
There are two simple and safe methods of using the API token 
without exposing it to unintended audiences.

#### Storing the token in a file and loading it as needed
Simply put your API token in a file on your harddrive and read
the file when initializing the Coscine client. This has the 
advantage of keeping the token out of the sourcecode and
offering the user an easy way to switch between tokens by
changing the filename in the sourcecode.
```python
import os

fd = open("token.txt", "rt")
token = fd.read()
fd.close()
```
However it comes at the disadvantage of potentially exposing
the token by accidentially leaking the file together with
the sourcecode e. g. on online platforms such as GitHub.
Therefore precautions must be taken i. e. when using git as a
versioning system. A `.gitignore` file including any possible
token name or file extension should be mandatory. You could
for example exclude the filename *token.txt*. A better way
would be to agree upon a common token file extension
such as `.token` and exclude that file extension. Then you can 
safely push your code to online platforms such as GitLab
or GitHub, since the `.gitignore` makes sure your sensitive
files are not included in any uploads.  
- [gitignore guide by W3schools.com](https://www.w3schools.com/git/git_ignore.asp)
- [gitignore docs on git-scm.com](https://git-scm.com/docs/gitignore)

#### Storing the token in an environment variable
This method does not rely on any files but instead on
[environment variables](https://en.wikipedia.org/wiki/Environment_variable).
Simply set an environment variable containing your token and
use that variable from your python program.
***Set environment variable:***  
```python
import os
os.environ["COSCINE_API_TOKEN"] = "My Token Value"
```
You would do this only once, and obviously not in your
final program/script. There are alternative
platform-specific methods for creating an environment 
variable (see links given below), but this pythonic
approach should work on most # platforms out of the box.
Just don't do this in your actual program - do it with an 
extra script or in an interactive python session (shell).
***Get the environment variable***
```python
import os
token = os.getenv("COSCINE_API_TOKEN")
```
This is certainly a little more complex for some users
who may want to use your program. They can easily share
tokens by sending a file to colleagues but sharing
environment variables requires each user to additionally
create the environment variable on their local system.  
Find out how to temporarily or permanently set environment
variables on certain Operating Systems:  
- [Windows](http://www.dowdandassociates.com/blog/content/howto-set-an-environment-variable-in-windows-command-line-and-registry/)
- [Linux](https://phoenixnap.com/kb/linux-set-environment-variable)
