# Using the Coscine Python SDK
## General Information
### Input Forms
In many places of the Python SDK we use a custom datatype called
InputForm, which is basically a dictionary with a lot of extra
functionality added on top (such as controlled vocabularies). They provide
an easier way of interacting with metadata by presenting and expecting
data in human-readable form and only internally generating the Coscine
expected format, which consists of a lot of ambiguous URLs and identifiers.  

#### Motivation
> Why does the Coscine Python SDK use InputForms over methods with
named arguments like the following one?
```python
def create_project("test", "test", datetime.now(), ...)
```
Multiple benefits arise from the use of a dictionary:
- The dictionary keys are familiar from the Coscine web interface and
  change based on the python client language preset
- The fields inside of the dictionary can be iterated over and their
  properties like controlled vocabularies can be queried

The 2nd benefit ultimately enables easy use in GUI applications.
You will encounter InputForms when dealing with
Project, Resource or FileObject metadata.

#### Printing an InputForm
InputForms present metadata in human-readable form and expect data in
human-readable format - just like in the Coscine web interface. When
printing an InputForm, required fields are marked with an asterisk (*)
next to their property name. Fields controlled by a vocabulary are marked with
an `X` in the C (=controlled) column. An example InputForm for metadata
of a FileObject is printed below.
```
+---+------------------+--------------------------------+
| C | Property         | Value                          |
+---+------------------+--------------------------------+
|   | Creator*         | Dr. Akula                      |
|   | Title*           | Pneumonia positive             |
|   | Production Date* | 2021-05-20                     |
| X | Subject Area     | Radiology and Nuclear Medicine |
| X | Resource         | Image                          |
|   | Rights           |                                |
|   | Rightsholder     |                                |
+---+------------------+--------------------------------+
```

#### Gaining access to the controlled vocabulary
Some fields inside an InputForm are controlld by a vocabulary. These fields
only accept values that are contained inside of their vocabulary. To find
out which values that includes, we can directly access the underlying
vocabulary with
```python
import coscine

client = coscine.Client("My Coscine API Token")
form = client.project_form()  # or any other form
# get the vocabulary of the Disciplines key
vocabulary = form.vocabulary("Discipline")

# Find out whether the vocabulary contains some stuff
if vocabulary.contains("Alienology"):
    send_greetings_to_mars("They know! Prepare the Invasion!")

# Look up the Coscine internal representation of a key
# We can also do the other direction and lookup the key given an internal
# value with vocabulary.lookup_value()
value = vocabulary.lookup_key("Computer Science 409")
print(value)  # yields:
# {'id': 'cfd6b656-f4ba-48c6-b5d8-2f98a7b60d1f', 'url': 'http://www.dfg.de/dfg_profil/gremien/fachkollegien/liste/index.jsp?id=409', 'displayNameDe': 'Informatik 409', 'displayNameEn': 'Computer Science 409'}

# Print all keys of the vocabulary which includes all accepted values
for key in vocabulary.keys():
    print(key)  # yields a lot of keys | Useful for GUI dropdown lists
```

## ... With Projects
Coscine is a project-based data management platform. A project represents
the central hub for all of the data resources and people involved
in a scientific undertaking.

### The Project object
The Coscine Python SDK abstract Coscine projects with a
custom Project datatype. An instance of type `coscine.Project`
has the following properties:

|Property               |   Type | Description |
|-----------------------|--------|-------------|
|id|`str` |Unqiue Project identifier|
|name|`str` |Full Project Name|
|display_name|`str` |Project Display Name|
|description|`str` |Project Description|
|principle_investigators|`str` |Project Principle Investigators|
|start_date|`datetime` |Project Inception date|
|end_date|`datetime` |End of project lifespan|
|disciplines|`list[str]` |List of involved scientific disciplines|
|organizations|`list[str]` |List of involved institutions|
|visibility|`str` |Project visiblity setting|

Properties can be accessed via dot notation
i. e. `print(project.name)`.

### Get a list of projects
The following snippet demonstrates how to get a list of
Coscine Projects and prints all returned projects by their
full name. The call `client.projects()` will only return
toplevel projects (i.e. only the projects at the root level).
To include subprojects in the results, the `toplevel`
parameter must be set to `False` as demonstrated in
another snippet further down below.
```python
import coscine

client = coscine.Client("My Coscine API Token")
projects = client.projects()
for project in projects:
	print(project.name)
```

#### Filtering the Project result list
The resulting list of projects can be filtered according
to the project properties:
```python
from datetime import datetime
import coscine

client = coscine.Client("My Coscine API Token")

# Get a list of all projects that have already ended.
# This time including subprojects (toplevel=False).
projects = client.projects(toplevel=False)
results = filter(lambda project: project.end_date < datetime.now(), projects)
for project in list(results):
	print(f"Project {project.name} has already ended.")
```

#### Getting an individual project
Individual projects can directly be accessed through
the `client.project()` method. This is just a convenience
function, that internally does the same filtering as
described above. It does however not return a list - it
guarantees to return a single Project or fail.
```python
try:
	project = client.project("My Project Display Name")
	print(project.name)
except IndexError:
	print("More than 1 project with the same name!")
```

#### Dealing with Subprojects
The project object itself contains methods to
query subprojects. Those are equivalent to the methods
described above, but only return subprojects of a specific
project.
```python
project.subprojects()
project.subproject("Subproject Display Name")
```

### Printing a project
When printing a project object, its metadata is printed
formatted inside of a table.  

```python
print(project)
```
*yields:*
```
+----------------------------------------------------------------+
|                         Project [Cool]                         |
+-------------------------+--------------------------------------+
|         Property        |                Value                 |
+-------------------------+--------------------------------------+
|            ID           | 501cfbe5-39ca-414e-9b42-2b033d740eef |
|           Name          |                 Cool                 |
|       Display Name      |                 Cool                 |
| Principle Investigators |              DFG Vampir              |
|       Disciplines       |         Computer Science 409         |
|      Organizations      |        https://ror.org/04xfq0f34     |
|        Start Date       |         2022-10-29 22:10:05          |
|         End Date        |         2022-10-29 22:10:05          |
|        Visibility       |           Project Members            |
+-------------------------+--------------------------------------+
```

The keys and values are human readable and easy to understand.
Under the hood it does not look that simple but sometimes we might
need to access the Coscine internal identifiers for a certain field.
In that case we can use the *data* dictionary, which grants us
access to the metadata as seen and expected by the Coscine servers.
Printing it yields a JSON representation of our projects metadata:

```python
print(project.data)
```
*yields:*

```
{
    "projectName": "Cool",
    "displayName": "Cool",
    "description": "Blablabla",
    "principleInvestigators": "DFG Vampir",
    "startDate": "2022-10-29T22:10:05",
    "endDate": "2022-10-29T22:10:05",
    "disciplines": [
        {
            "id": "cfd6b656-f4ba-48c6-b5d8-2f98a7b60d1f",
            "url": "http://www.dfg.de/dfg_profil/gremien/fachkollegien/liste/index.jsp?id=409",
            "displayNameDe": "Informatik 409",
            "displayNameEn": "Computer Science 409"
        }
    ],
    "organizations": [
        {
            "displayName": "RWTH Aachen University",
            "url": "https://ror.org/04xfq0f34"
        }
    ],
    "visibility": {
        "id": "8ab9c883-eb0d-4402-aaad-2e4007badce6",
        "displayName": "Project Members"
    }
}
```
The data looks different and is (subjectively) a lot harder to interact
with than before. Coscine does not care about human readable values
such as `RWTH Aachen University` - those are actually silently ignored.
What really matters are the unique identifiers and urls. Working with
those may lead to one or another headache though... Thankfully the
*ProjectForm* abstracts all of that away by assigning these identifiers
automatically with its `generate()` method.
The *ProjectForm* is described in detail in the example code for creating
a project.

### Create a project
The python module provides a simple means of creating Coscine Projects:
The *ProjectForm*.

#### The Project Form

#### Preset initialization
```python
import coscine

client = coscine.Client("My Coscine API Token")
project_metadata: dict = {
	"Project Name": "My Name",
	"Display Name": "My Display Name",
	...
	"Visibility": "Project Members"
}
form = client.project_form(data)
client.create_project(form)
## Alternatively:
# form = client.project_form()
# form.fill(data)
```
#### Manual initialization/editing
```python
import coscine

client = coscine.Client("My Coscine API Token")
form = client.project_form()
form["Projektname"] = "test"
form["Anzeigename"] = "test"
form["Projektbeschreibung"] = "test"
form["Principal Investigators"] = "Mr. Test"
form["Projektstart"] = datetime.now()
form["Projektende"] = datetime.now()
form["Disziplin"] = ["Informatik 409"]
form["Teilnehmende Organisation"] = ["https://ror.org/04xfq0f34"]
form["Sichtbarkeit"] = "Project Members"
client.create_project(form)
```
>**Special consideration** must be taken for the Organizations field!  
You can't just enter "RWTH Aachen" or similar organizations here, you have to
first look them up at "[https://ror.org/search](https://ror.org/search)". If
you do just put in an invalid string, you break something in your project.
The creation will not fail, but you may get an error when visiting the
project settings. Do not worry - you can simply fix the issue later on via
the web interface or by updating the project metadata with the python sdk.  
The reason for all these shenanigans it, that the organizations database
exceeds a couple of gigabytes. We cannot use a controlled vocabulary for this.
You can query ror urls yourself and search for organizations via the ror API:
[https://ror.readme.io/docs/rest-api](https://ror.readme.io/docs/rest-api).

Just like any other `coscine.InputForm` you can `print()` a form to
get a human-readable overview over the data.

### Delete a project
If we create projects with the Coscine Python SDK we may also need to
delete some projects.
```python
import coscine

client = coscine.Client("My Coscine API Token")
project = client.project("Some Project I don't like")
try:
    project.delete()
except PermissionError:
    print("We are not authorized to delete that project. :(")
```
Be aware though, that this function may fail due to unsufficient privileges.
Only project owners can delete a project - members can not!
The call may yield an error, which we should catch and handle.

### Download a project
Downloading a project and all of the resources inside of that project
is a piece of cake with the `download()` method. It accepts an optional
path where it should save the project to. If no path is specified, the
current local directory is chosen as path.
```python
import coscine

client = coscine.Client("My Coscine API Token")
project = client.project("Fancy project - please don't steal")
project.download(path="./")  # This is my project now huehehe
```

### Member management
Coscine Projects contain Project Members. We can get a list of the members
of a project by calling the `members()` method.
```python
import coscine

client = coscine.Client("My Coscine API Token")
project = client.project("My Project with Members")
members = project.members()
for member in members:
    print(member.name)
    print(member.email)

# Set project role for a member
member.set_role("Owner")
    
# Remove a member from the project
member.remove()
```
Inviting new members:
```python
EMAIL_ADDRESS: str = "john@example.com"
ROLE: str = "Member"  # or "Owner"
project.invite(EMAIL_ADDRESS, ROLE)
```

## ... With Resources
In Coscine, Resources store all of your data and metadata. As such they
represent a key data structure, which you most certainly will
interact a lot with.

### The Resource object
The Coscine Python SDK abstract Coscine resources with a
custom Resource datatype. An instance of type `coscine.Resource`
has the following properties:

|Property               |   Type | Description |
|-----------------------|--------|-------------|
|id|`str` |Unqiue Resource identifier|
|pid|`str` |Persistent identifier of the Resource|
|name|`str` |Full Resource Name|
|display_name|`str` |Resource Display Name|
|license|`str` |Name of the license for the data contained in the resource|
|type|`str` |Resource Type, e. g. "rdss3rwth"|
|disciplines|`list[str]` |List of involved scientific disciplines|
|profile|`str` |Name of the Application Profile used in the Resource|
|archived|`bool` |Indicating whether the resource is archived (read-only)|
|creator|`str` |Resource Creator GUID|

Properties can be accessed via dot notation
i. e. `print(resource.name)`.

### Getting a list of resources
Analoguous to getting a list of projects, you can get a list
of resoures. The only difference is, that the `resources()`
method is part of the project object and does only query
resources contained within that project. Just like with
projects, you can filter by certain resource properties.

```python
project = client.project("Some project")
resources = project.resources()
for resource in resources:
    print(resource.name)
```

If we know which resource we require beforehand, we can just
query it by its name. We now get a single object of type
Resource, instead of a list:
```python
resource = project.resource("ResourceNameS3")
print(resource)
```

### Creating a resource
Once again we are using InputForms to set metadata of a Coscine object.
```python
form = client.resource_form()
form["Resource Type"] = "rds"
form["Resource Size"] = 31
form["Resource Name"] = "My Cool Resource"
form["Display Name"] = "Cool"
form["Resource Description"] = "Testing Coscine Client Resources"
form["Discipline"] = ["Computer Science 409"]
form["Application Profiles"] = "RADAR"
form["Visibility"] = "Project Members"
project.create_resource(form)
```

### Deleting a resource
```python
try:
    resource.delete()
except PermissionError:
    print("Not authorized (not a project owner).")
```

### Downloading a resource
Downloading a resource and all of the data contained within the resource is just as simple as downloading a project. In fact internally `project.download()` just calls `Resource.download()` for all resources contained within the project.

```python
resource.download(path="./")
```

### Resource Quota
You can fetch the used up quota of a resource as an integer indicating the size used in Bytes.

```python
quota = resource.quota()
print(quota)
```

### Resource Application profile
An application profile specifies a template for metadata. There may be times where you need to interact with that profile. To get the application profile of a resource you simple call the `application_profile()` method.
```python
# Print the application profile used by the resource
profile = resource.application_profile()
print(profile)
```

### Accessing S3 credentials
RDS-S3 resources can be directly accessed via an S3-client. Direct connections require S3 credentials, which s3 resource instances happily provide to us. Only works for s3 resources.

```python
# Keys with write privileges, substitute for key_read for read-only privileges
access_key: str = resource.s3.access_key_write
secret_key: str = resource.s3.secret_key_write
endpoint: str = resource.s3.endpoint
bucket: str = resource.s3.bucket
```

### Interfacing with S3 natively
The s3 instance of a resource provides some additional functionality when
the `boto3` package is installed, to enable users to directly
interface with S3.

You can create folders (though S3 only mimicks folders):
```python
resource.s3.mkdir("directory/")
```

And upload files without metadata and size restrictions:
```python
resource.s3.upload("root/Testfile.txt", "./myfile.txt")
```

### Resource to RDF graph
Coscine Resources contain FileObjects and Metadata. We can generate
an RDF graph from the metadata of files in a resource. With such
a graph a whole lot of new possibilities open up for interacting
with the data. We can run SPARQL queries on the triples stored
in the graph, serialize it into many different formats such as
XML, CSV, Turtle, [...] and in general use our data in line
with the concept of Semantic Web.  

#### Generating an RDF graph from a resource
The Coscine Python SDK internally uses rdflib for working with RDF
graphs. The `Resource.graph()` method returns an instance of type
`class rdflib.Graph`. We can then interact with that graph just
like with any other rdflib graph.  
Creating a graph for a lot of metadata can be a time consuming task.
If the verbose setting of the client is enabled, a progress bar
is shown in the command line to indicate the progress.
Anything over 1000 files will take a noticable amount
of time (seconds). Luckily, once generated, RDF graphs can be
stored on the harddisk and reloaded later on with minimal overhead.  
Only files with metadata are reflected in the RDF graph, other
files or folders are ignored, as they do not contribute relevant
information to the graph.
```python
resource = client.project("My Project").resource("My Resource")
graph = resource.graph()
```

#### Combining graphs of different resources
The RDF graph returned created from a resource can be merged with
RDF graphs of other resources, making it possible to perform
multi-resource SPARQL queries on your data and to visualize the
data relationship between resources. Intuitively it makes sense
that all resources should roughly have equivalent application
profiles and ontologies. It is not relevant whether the resources
lay within the same project.
```python
import rdflib
g1 = client.project("My Project").resource("My Resource 1").graph()
g2 = client.project("My Project").resource("My Resource 2").graph()
g3 = client.project("Different Project").resource("Resource").graph()
master_graph = g1 + g2 + g3
```

#### Serializing an RDF graph
RDF graphs can be serialized into various formats. For example
to XML:
```python
graph = client.project("P").resource("R").graph()
# Write the graph in XML format to 'file.xml'
graph.serialize("file.xml", format="xml")
# Our metadata is now accessible as XML in 'file.xml'!
```

#### Printing an RDF graph
The Coscine Python SDK ships with a small utility function to
serialize an RDF graph into dot format. Dot format can be used by
tools like GraphViz to generate nice looking figures, indicating
the relations between your data.  
The following snippet demonstrates how to create such a dotfile...
```python
resource = client.project("My Project").resource("My Resource")
graph = resource.graph()
with open("dotfile.dot", "wt", encoding="utf-8") as fd:
	coscine.utils.rdf_to_dot(graph, fd)
```
...which can subsequently be converted to an image with graphviz:
```
Graphviz/bin/dot -Tsvg dotfile.dot > image.svg
Graphviz/bin/fdp -Tsvg dotfile.dot > image2.svg
```
Graphviz offers multiple rendering engines (e.g. 'dot', 'fdp', ...),
each producing slightly different output.
It is worth checking them all out to figure out which suits
your data best.

## ... With File Objects
Files are stored inside of resources. However Resources do not necessarily contain files. It depends on the resource type. RDS and RDS-S3 contain files, but Linked Data resources contain references to files. Therefore we cannot just talk about files, but have to use a more abstract term such as 'object'. Objects represent files and file-like instances in Coscine.
Nonetheless the methods of interacting with files and file-like objects is always the same - just don't expect file contents for objects of linked data resources, as those merely contain links or whatever has been specified as their content.

### The FileObject Object
The Coscine Python SDK abstract Coscine files with a
custom FileObject datatype. An instance of type `coscine.FileObject`
has the following properties:

|Property               |   Type | Description |
|-----------------------|--------|-------------|
|name|`str` |Full Resource Name|
|size|`int` |Filesize in bytes|
|type|`str` |Type of FileObject, i. e. "file" or "folder"|
|modified|`datetime` |Last time the FileObject was modified|
|created|`datetime` |Time the FileObject was created|
|provider|`str` |Resource type|
|filetype|`str` |FileObject file type, e. g. '.png'|
|path|`str` |Full path including folders (in case of S3 resources)|
|is_folder|`bool` |Indicates whether the FileObject is a folder|
|has_metadata|`bool` |Indicates whether the FileObject has metadata|

Properties can be accessed via dot notation
i. e. `print(file.name)`.

### Getting a list of files and folders
```python
files = resource.objects()
for file in files:
    print(file.name)

# Files inside Folder/
files2 = resource.objects("Folder/")
```

```python
file = resource.object("Messung (1).bin")
print(file.name)
data = file.content()
```

### Getting all objects contained in a resource
The methods for files and folders only work on one level
of a directory hierarchy. To get all files and folders
regardless of hierarchy, we can use the `content()s` method:
```python
# No directory hierarchy anymore, all files on a flat level
all_files = resource.contents()
```

### Downloading a file
```python
file = resource.object("Messung (1).bin")
path = "./"
file.download(path)
```

### Uploading a file
```python
metadata = resource.metadata_form()

metadata["Title"] = "..."
# fill in fields

filename = "messung.bin" # filename as it should appear in Coscine
path = "./data/messung.bin" # path to file on harddrive
resource.upload(filename, path, metadata)
```

### Deleting a file
```python
file = resource.object("Messung (1).bin")
file.delete()
# The file object is still valid until garbage collected.
# The file on the Coscine server has already been removed though.
```

### Updating file metadata
We can interact with metadata using a MetadataForm.
```python
form = file.form()
print(form)
```

The form fields change depending on the selected application profile for the resource.
```python
form["Title"] = "My Title"

# Update metadata
file.update(form)
```
