###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
###############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file contains some tests to make sure that everything in `src/coscine`
works as intended.
For testing we are using the python unittest test suite, however due to
many of our functions leading to (expected) side-effects, we "abuse"
the unittest package a little bit by not sticking to general unit testing
guidelines. That means we are not mocking requests or anything like that -
everything is performed on the live system.
"""

###############################################################################
# Dependencies
###############################################################################

from datetime import datetime
import unittest
import os
import os.path
import sys
import logging
sys.path.append("src")
import coscine

###############################################################################
# Configuration
###############################################################################

# Set via environment variables
# -> Makes it easy to set token in gitlab ci pipeline
TOKEN: str = os.getenv("COSCINE_API_TOKEN")

# Master project display name
# -> This project must already be present and have sufficient quota
# for all sorts of resource types
MASTER_PROJECT_NAME: str = "Solaris"

# Path to the tests directory
BASEPATH: str = os.path.dirname(os.path.realpath(__file__))

# Set up logger
logging.basicConfig(level=logging.INFO)

###############################################################################
# Class definitions / Functions / Scripts
###############################################################################

class CoscineUnittest(unittest.TestCase):
	client: coscine.Client
	master_project: coscine.Project

###############################################################################

	@classmethod
	def setUpClass(cls) -> None:
		cls.client = coscine.Client(TOKEN)
		cls.master_project = cls.client.project(MASTER_PROJECT_NAME)
		return super().setUpClass()

###############################################################################

	def test_switch_language(self) -> None:
		self.client.settings.language = "de"
		self.client.settings.language = "en"
		with self.assertRaises(ValueError):
			self.client.settings.language = "es"

###############################################################################

	def test_project_form_fill(self) -> None:
		data: dict = {
			"Project Name": "Test Name",
			"Display Name": "Test Display Name",
			"Project Description": "Test Description",
			"Principal Investigators": "Test Principal Investigator",
			"Project Start": datetime.now(),
			"Project End": datetime.now(),
			"Discipline": ["Computer Science 409"],
			"Participating Organizations": ["https://ror.org/04xfq0f34"],
			"Project Keywords": "Test Keywords",
			"Metadata Visibility": "Project Members",
			"Grant ID": "Test Grant ID"
		}
		form = self.client.project_form(data)
		self.assertIsInstance(form, coscine.ProjectForm)
		self.assertIsInstance(form.generate(), dict)

###############################################################################

	def test_resource_form_fill(self) -> None:
		data: dict = {
			"Resource Type": "rdss3rwth",
			"Resource Size": 16,
			"Resource Name": "Test Name",
			"Display Name": "Test Display Name",
			"Resource Description": "Test Description",
			"Discipline": ["Computer Science 409", "Mathematics 312"],
			"Resource Keywords": "Test Keywords",
			"Metadata Visibility": "Project Members",
			"License": "MIT License",
			"Internal Rules for Reuse": "Test Rules",
			"Application Profile": "engmeta"
		}
		form = self.client.resource_form(data)
		self.assertIsInstance(form, coscine.ResourceForm)
		self.assertIsInstance(form.generate(), dict)

###############################################################################

	def test_project_form_parse(self) -> None:
		form = self.master_project.form()
		self.assertIsInstance(form, coscine.ProjectForm)

###############################################################################

	def test_list_project_members(self) -> None:
		# There is always at least one project member per project
		for member in self.master_project.members():
			self.assertIsInstance(member, coscine.ProjectMember)

###############################################################################

###############################################################################

	def test_subproject(self) -> None:
		projectForm = self.client.project_form()
		self.assertIsInstance(projectForm, coscine.ProjectForm)
		projectForm["Project Name"] = "coscine-sdk-test-project"
		projectForm["Display Name"] = "sdk-test-project"
		projectForm["Project Description"] = (
			"A test project for testing the Coscine Python SDK"
		)
		projectForm["Principal Investigators"] = "Lt. Cmndr. Research Data"
		projectForm["Project Start"] = datetime.now()
		projectForm["Project End"] = datetime.now()
		projectForm["Discipline"] = "Computer Science 409"
		projectForm["Participating Organizations"] = "https://ror.org/04xfq0f34"
		projectForm["Metadata Visibility"] = "Project Members"
		subproject = self.master_project.create_subproject(projectForm)
		self.assertIsInstance(subproject, coscine.Project)

		# Edit project metadata
		project = self.client.project(MASTER_PROJECT_NAME)\
							.subproject("sdk-test-project")
		projectForm = project.form()
		projectForm["Display Name"] = "edited-test-project"
		project.update(projectForm)

		# Clean up
		subproject.delete()

###############################################################################

	def test_s3_resource(self) -> None:
		# Create an S3-resource
		resourceForm = self.client.resource_form()
		self.assertIsInstance(resourceForm, coscine.ResourceForm)
		resourceForm["Resource Type"] = "rdss3rwth"
		resourceForm["Resource Size"] = 1
		resourceForm["Resource Name"] = "test-resource-s3"
		resourceForm["Display Name"] = "test-resource-s3"
		resourceForm["Resource Description"] = "test-resource-s3"
		resourceForm["Discipline"] = ["Computer Science 409", "Mathematics 312"]
		resourceForm["Metadata Visibility"] = "Project Members"
		resourceForm["Application Profile"] = "radar"
		s3Resource = self.master_project.create_resource(resourceForm)
		self.assertIsInstance(s3Resource, coscine.Resource)

		# Create folders and files inside of the S3 resource
		metadataForm = s3Resource.metadata_form()
		metadataForm["Title"] = "Hello World!"
		metadataForm["Creator"] = "John Doe"
		metadataForm["Production Date"] = datetime.now()
		metadataForm["Resource"] = "Dataset"
		metadataForm["Subject Area"] = "History"
		s3Resource.upload("root/child/leaf/testfile.txt",
			os.path.join(BASEPATH, "data/testfile.txt"), metadataForm)
		metadataForm["Title"] = "Flohzirkus"
		s3Resource.upload("Lumpi.JPG", os.path.join(BASEPATH,
				"data/testfile.txt"), metadataForm.generate())

		# Edit resource metadata
		resource = self.master_project.resource("test-resource-s3")
		print(resource)
		resourceForm = resource.form()
		resourceForm["Display Name"] = "edited-resource-name"
		print(resource.update(resourceForm))

		# Edit file metadata
		object = resource.object("root/child/leaf/testfile.txt")
		metadataForm = object.form()
		metadataForm["Title"] = "John Doe ♥ Jane"
		object.update(metadataForm)

		# Clean up
		s3Resource.delete()

###############################################################################

if __name__ == "__main__":
	unittest.main()

###############################################################################
